﻿using System;
using System.Linq;
using System.Threading.Tasks;
using ClientServer.Controllers.Core.Exercises.DoExerciseAfterSolution;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models.Exercises;
using ClientServer.Models.Exercises.Release;
using ClientServer.Models.Exercises.Solution;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;

namespace ClientServer.Controllers.Core.Exercises
{
    /// <summary>
    /// controller for the tutor view submission page
    /// 
    /// we need to load data in 2 steps because we cannot load the assessment directly in case there is no assessment yet
    /// so we load everything that is possible (solution, tests, exercise, ....)
    /// then try to load the assessment
    /// this controller should be able to create an assessment if there is no assessment yet (tutor can give manual points...)
    /// </summary>
    [Route(Constants.ApiPrefix + "tutor/view/assessment")]
    public class TutorViewAssessmentController : ControllerWithDb
    {
        public TutorViewAssessmentController(YapexDbContext context) : base(context)
        {
        }


        private async Task<bool> HasPermission(int exerciseReleaseId)
        {
//            var exists = await _context.ExerciseReleases.CountAsync(p => p.Id == exerciseReleaseId);
//
//            if (exists == 0) {
//                await
//                    Response.WriteAsync(
//                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "release not found")));
//                return false;
//            }
            var targetUserGroupId =
                await _context.ExerciseReleases
                    .Where(p => p.Id == exerciseReleaseId)
                    .Select(p => p.Exercise.UserGroupId)
                    .FirstOrDefaultAsync();


            if (!await base.HasGroupPermission(targetUserGroupId,
                permission => permission != null && permission.CanAssessExercises))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return false;
            }

            return true;
        }

        internal async Task<ExerciseRelease> _GetExerciseRelease(int exerciseReleaseId)
        {
            var exerciseRelease = await _context.ExerciseReleases
                .FirstOrDefaultAsync(p => p.Id == exerciseReleaseId);

            if (exerciseRelease == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "exercise release not found")));
                return null;
            }

            return exerciseRelease;
        }

        /// <summary>
        /// returns everything for the page except the assessment
        /// exercise (with description, tests, ... ...)
        /// user solution
        /// user custom tests
        /// test results...
        /// </summary>
        /// <param name="userId">the target user id</param>
        /// <param name="pLangId"></param>
        /// <param name="exerciseReleaseId"></param>
        [HttpGet("exercise/{exerciseReleaseId}/{userId}/{pLangId}")]
        public async Task GetExercise(int exerciseReleaseId, int userId, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;

            int ownUserId = GetUserId();


            if (!await HasPermission(exerciseReleaseId)) return;

            //var targetUserGroup =
            //    _context.ExerciseReleases.Where(p => p.Id == exerciseReleaseId)
            //        .Select(p => p.Exercise.UserGroupId)
            //        .FirstOrDefault();

            //if (!base.HasGroupPermission(targetUserGroup,
            //    permission => permission != null && permission.CanAssessExercises))
            //{
            //    await
            //        Response.WriteAsync(
            //            Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
            //    return;
            //}


            var userSolution = await _context.Solutions
                .Include(p => p.SolutionFiles)
                .FirstOrDefaultAsync(
                    p => p.ExerciseReleaseId == exerciseReleaseId && p.UserId == userId && p.PLangId == pLangId);


            if (userSolution == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "solution not found")));
                return;
            }

            //now we need an exercise like in the do exercise controller...

            //we only need to exercise and custom tests... for the solution files we use the do exercise controller...

            //TODO merge this with the do exercise controller
            //e.g. extract method for just getting the exercise ....

            //we dont need the templates

            //for the test results we can use to do exercise func GetAllTestResults

            ExerciseRelease exerciseRelease = await _context.ExerciseReleases
                    .Where(p => p.Id == exerciseReleaseId)
                    .Include(p => p.Exercise)
                    .Include(p => p.PLang)

                    //test cases
                    .Include(p => p.Exercise.Tests)
                    .ThenInclude(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceTestAsset)
                    .Include(p => p.Exercise.Tests)
                    .ThenInclude(p => p.TestSettings)
                    .Include(p => p.Exercise.Tests)
                    .ThenInclude(p => p.TestType)
                    .Include(p => p.Exercise.CodeTemplates)
                    .ThenInclude(p => p.TemplateFiles)

                    //we don't need code templates here

                    //task description
                    .Include(p => p.Exercise.Description)
                    .ThenInclude(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceMarkdownAsset)
                    .FirstOrDefaultAsync()
                ;


            if (exerciseRelease?.Exercise == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise or release not found")));
                return;
            }

            Exercise exercise = exerciseRelease.Exercise;


            ExerciseReleaseWithUserAsParticipation oldParticipation = await _context
                    .ExerciseReleaseWithUserAsParticipations
                    .Include(p => p.CustomTests)
                    .ThenInclude(p => p.TestResultsNew)
                    .FirstOrDefaultAsync(p => p.ExerciseReleaseId == exerciseRelease.Id && p.UserId == userId)
                ;

            if (oldParticipation == null)
            {
                //user has no old participation so he/she never did this exercise... possible?
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }

            #region forge exercise for frontend

            var filesIdsToRead = exercise.Description.AssetReferences
                .Select(p => p.FileReferenceMarkdownAssetId)
                .ToArray();

            var basePath = Files.GetUploadFilePath(UploadDirType.MarkdownAssets);

            var filesDict = await Files.ReadUploadedFiles(basePath, filesIdsToRead);
            
            var exerciseForFrontend = new DoExerciseFromBackend
            {
                Id = exercise.Id,
                CanUserCreateFiles = exercise.CanUserCreateFiles,
                CanUserCreateCustomTests = exercise.CanUserCreateCustomTests,
                IsPermanentlyLocked = exercise.IsPermanentlyLocked,
                DisplayName = exercise.DisplayName,
                ShortDescription = exercise.ShortDescription,
                CodeTemplates = exercise.CodeTemplates.Select(p => new ExerciseCodeTemplateFullBase()
                {
                    //TODO maybe not send all template...make separate request or with solution because then we know the plang?
                    Id = p.Id,
                    MainFileId = p.MainFileId ?? 0,
                    PLangId = p.PLangId,
                    TemplateFiles = p.TemplateFiles.Select(t => new ExerciseCodeTemplateFileFullBase()
                    {
                        Id = t.Id,
                        Content = t.Content,
                        DisplayIndex = t.DisplayIndex,
                        FileNameWithExtension = t.FileNameWithExtension,
                        IsEditableByUser = t.IsEditableByUser
                    }).ToList()
                }).ToList(),
                ExerciseRelease = new ExerciseReleaseDoExerciseFromBackend
                {
                    Id = exerciseRelease.Id,
                    AvailableWorkingTimeInMinutes = exerciseRelease.AvailableWorkingTimeInMinutes,
                    HasLimitedWorkingTime = exerciseRelease.HasLimitedWorkingTime,
                    PLangId = exerciseRelease.PLangId,
                    ReleaseCode = exerciseRelease.GeneratedCode,
                    IsReleased = exerciseRelease.IsReleased,
                    AutomaticEndAt =
                        exerciseRelease.ReleaseDurationType == ReleaseDurationType.Infinitely
                            ? null
                            : exerciseRelease.AutomaticEndAt,
                    HasAutomaticAssessmentFinished = exerciseRelease.HasAutomaticAssessmentFinished,
                    ReleaseDurationType = (int) exerciseRelease.ReleaseDurationType,
                    ReleaseStartType = (int) exerciseRelease.ReleaseStartType,
                    ShouldClearClipboard = exerciseRelease.ShouldClearClipboard,
                    HideSiteHeaderBar = exerciseRelease.HideSiteHeaderBar,
                    HideExerciseLeaveActions = exerciseRelease.HideExerciseLeaveActions,
                    HidePrintOptions = exerciseRelease.HidePrintOptions,
                    ShowAdditionalLogButton = exerciseRelease.ShowAdditionalLogButton,
                    MaxManualPoints = exerciseRelease.MaxManualPoint,
                },
                ExerciseDescription = new ExerciseDescriptionFullBase
                {
                    Id = exercise.Description.Id,
                    Content = exercise.Description.Content,
                    Assets =
                        exercise.Description.AssetReferences.Select(
                            asset => new FileWithData
                            {
                                Id = asset.FileReferenceMarkdownAsset.Id,
                                DisplayName = asset.FileReferenceMarkdownAsset.OriginalName,
                                MimeType = asset.FileReferenceMarkdownAsset.MimeType,
                                Content = filesDict[asset.FileReferenceMarkdownAssetId],
                            }).ToList()
                }
            };

            #endregion

            #region tests

            foreach (var test in exercise.Tests)
            {
                //we load the submit tests again in frontend (when we get the results)
                //and use the submit tests with the results not those...
                //so we only need the normal tests here!
                if (test.IsSubmitTest == false)
                {
                    exerciseForFrontend.Tests.Add(new ExerciseTestFromBackend
                    {
                        Id = test.Id,
                        DisplayName = test.DisplayName,
                        Content = test.Content,
                        Weight = test.Weight,
                        TestTypeId = test.TestType.Id,
                        TestSettings = new TestSettingsFullBase
                        {
                            Id = test.TestSettings.Id,
                            MemoryLimitInKb = test.TestSettings.MemoryLimitInKb,
                            TimeoutInMs = test.TestSettings.TimeoutInMs,
                            MaxDiskSpaceInKb = test.TestSettings.MaxDiskSpaceInKb,
                            CompileTimeoutInMs = test.TestSettings.CompileTimeoutInMs,
                            CompilerOptions = test.TestSettings.CompilerOptions,
                        },
                        Files = test.AssetReferences.Select(p => new FilePreviewFromBackend()
                        {
                            Id = p.FileReferenceTestAsset.Id,
                            OriginalName = p.FileReferenceTestAsset.OriginalName,
                            MimeType = p.FileReferenceTestAsset.MimeType,
                            SizeInBytes= p.FileReferenceTestAsset.SizeInBytes,
                        }).ToList(),
                        IsSubmitTest = false, //we never give the student the submit tests
                        DisplayIndex = test.DisplayIndex
                    });
                }
            }

            #endregion

            //get custom tests if any

            //tutor don't need to get the users custom tests
//            #region custom tests (user tests)
//
//            //because we copy pasted from do exercise controller and we changed the oldParticipation == null check...
//            if (oldParticipation != null) 
//            {
//                exerciseForFrontend.CustomTests = new List<CustomTestFromBackend>();
//
//                foreach (CustomTest oldCustomTest in oldParticipation.CustomTests)
//                {
//                    var testForBackend = new CustomTestFromBackend()
//                    {
//                        Id = oldCustomTest.Id,
//                        DisplayName = oldCustomTest.DisplayName,
//                        Content = oldCustomTest.Content,
//                        TestTypeId = oldCustomTest.TestTypeId,
//                        DisplayIndex = oldCustomTest.DisplayIndex,
//                        Files = oldCustomTest.TestAssets.Select(asset => new TestAssetFromBackendWithData()
//                        {
//                            //do not include content here ... same reason as the normal tests even if the user knows the content...
//                            Id = asset.Id,
//                            DisplayName = asset.DisplayName,
//                            MimeType = asset.MimeType,
//                            Size = asset.Content.Length,
//                            Content = null,
//                            Hash = asset.Hash
//                        }).ToList(),
//                    };
//
//                    //do not include custom test results here because they are plang dependent (include when getting user solution)
//
//                    exerciseForFrontend.CustomTests.Add(testForBackend);
//                }
//            }
//
//            #endregion


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<DoExerciseFromBackend>(ResponseCode.Ok,
                        "do exercise queried",
                        exerciseForFrontend)));
        }


        //--- user solution

        [HttpGet("solution/{exerciseReleaseId}/{userId}/{pLangId}")]
        public async Task GetSolutionForExercise(int exerciseReleaseId, int userId, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;

            if (!await HasPermission(exerciseReleaseId)) return;

            var exerciseRelease = await _GetExerciseRelease(exerciseReleaseId);
            if (exerciseRelease == null) return;

            var doExerciseController = new DoExerciseController(_context) {ControllerContext = this.ControllerContext};
            await doExerciseController._GetSolutionForExercise(exerciseRelease.GeneratedCode, pLangId, userId);
        }

        [HttpGet("testresults/{exerciseReleaseId}/{userId}/{pLangId}")]
        public async Task GetAllTestResults(int exerciseReleaseId, int userId, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;

            if (!await HasPermission(exerciseReleaseId)) return;

            var exerciseRelease = await _GetExerciseRelease(exerciseReleaseId);
            if (exerciseRelease == null) return;

            var doExerciseController = new DoExerciseController(_context) {ControllerContext = this.ControllerContext};
            await doExerciseController._GetAllTestResults(exerciseRelease.GeneratedCode, pLangId, userId);
        }

        [HttpGet("testresults/submit/{exerciseReleaseId}/{userId}/{pLangId}")]
        public async Task GetSubmitTestsWithResults(int exerciseReleaseId, int userId, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;

            if (!await HasPermission(exerciseReleaseId)) return;

            var exerciseRelease = await _GetExerciseRelease(exerciseReleaseId);
            if (exerciseRelease == null) return;

            var doExerciseController = new DoExerciseController(_context) {ControllerContext = this.ControllerContext};
            await doExerciseController._GetSubmitTestWithResultsAndAssessment(exerciseRelease.GeneratedCode, pLangId,
                userId);
        }

        //--- after solution


        [HttpGet("after/solution/{exerciseReleaseId}/{userId}/{pLangId}")]
        public async Task GetAfterSolutionForExercise(int exerciseReleaseId, int userId, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;

            if (!await HasPermission(exerciseReleaseId)) return;

            var exerciseRelease = await _GetExerciseRelease(exerciseReleaseId);
            if (exerciseRelease == null) return;


            var doExerciseAfterSolutionController = new DoExerciseAfterSolutionController(_context)
                {ControllerContext = this.ControllerContext};
            await doExerciseAfterSolutionController._GetAfterSolution(exerciseRelease.GeneratedCode, pLangId, userId);
        }

        [HttpGet("after/testresults/{exerciseReleaseId}/{userId}/{pLangId}")]
        public async Task GetAllAfterTestResults(int exerciseReleaseId, int userId, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;

            if (!await HasPermission(exerciseReleaseId)) return;

            var exerciseRelease = await _GetExerciseRelease(exerciseReleaseId);
            if (exerciseRelease == null) return;

            var doExerciseController = new DoExerciseAfterSolutionController(_context)
                {ControllerContext = this.ControllerContext};
            await doExerciseController._GetAllAfterTestResults(exerciseRelease.GeneratedCode, pLangId, userId);
        }

        [HttpGet("after/testresults/submit/{exerciseReleaseId}/{userId}/{pLangId}")]
        public async Task GetAfterSubmitTestWithResults(int exerciseReleaseId, int userId, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;

            if (!await HasPermission(exerciseReleaseId)) return;

            var exerciseRelease = await _GetExerciseRelease(exerciseReleaseId);
            if (exerciseRelease == null) return;

            var doExerciseController = new DoExerciseAfterSolutionController(_context)
                {ControllerContext = this.ControllerContext};
            await doExerciseController._GetAllSubmitTestsAndResults(exerciseRelease.GeneratedCode, pLangId, userId);
        }

        //--- assessment

        /// <summary>
        /// returns the assessment for a user for a release & plang
        /// returns an assessment even if there is no automatic assessment yet (we return some basic data e.g. name, plang, release, ...)
        /// </summary>
        [HttpGet("assessment/{exerciseReleaseId}/{userId}/{pLangId}")]
        public async Task GetAssessment(int exerciseReleaseId, int userId, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;
            
            if (!await HasPermission(exerciseReleaseId)) return;

            //get all automatic assessments for this release
            var release = await _context.ExerciseReleases
                    .Include(p => p.Exercise)
                    .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                    .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                    .FirstOrDefaultAsync(p => p.Id == exerciseReleaseId)
                ;

            if (release == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }

            var userReleaseParticipation = await _context.ExerciseReleaseWithUserAsParticipations.Where(p =>
                        p.ExerciseReleaseId == exerciseReleaseId
                        && p.UserId == userId
                    )
                    .Include(p => p.Solutions)
                    .ThenInclude(p => p.Assessment)
                    .Include(p => p.User)
                    .FirstOrDefaultAsync()
                ;

            if (userReleaseParticipation == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no participation found")));
                return;
            }

            var solution = userReleaseParticipation.Solutions.FirstOrDefault(p => p.PLangId == pLangId);

            if (solution == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                            "no solution for plang " + pLangId + " found")));
                return;
            }

            //assessment can be null if the assessment has not already finished ...
            //TODO maybe create one for manual assessment but then the automatic assessment should not clear the manual assessment

            var assessment = new AssessmentFullBase()
            {
                FirstName = userReleaseParticipation.User.FirstName,
                LastName = userReleaseParticipation.User.LastName,
                UserId = userReleaseParticipation.UserId,
                PLangId = solution.PLangId,
                ExerciseReleaseId = userReleaseParticipation.ExerciseReleaseId,
                SubmitTestPoints = solution.Assessment?.SubmitTestPoints,
                MaxSubmitTestPoints = solution.Assessment?.MaxSubmitTestPoints,
                NormalTestPoints = solution.Assessment?.NormalTestPoints,
                MaxNormalTestPoints = solution.Assessment?.MaxNormalTestPoints,
                ManualPoints = solution.Assessment?.ManualPoints,
                NoteForOtherTutors = solution.Assessment?.NoteForOtherTutors,
                FeedbackForStudent = solution.Assessment?.FeedbackForStudent,
                LastAssessmentErrorMessage = solution.Assessment?.FeedbackForStudent
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<AssessmentFullBase>(ResponseCode.Ok,
                        "", assessment)));
        }

        [HttpPost("assessment/update")]
        public async Task UpdateAssessment([FromBody] AssessmentFullBase assessmentFullBase)
        {
            if (!await base.IsLoggedIn()) return;

            
            
            //get all automatic assessments for this release
            var release = await _context.ExerciseReleases
                    .Include(p => p.Exercise)
                    .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                    .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                    .FirstOrDefaultAsync(p => p.Id == assessmentFullBase.ExerciseReleaseId)
                ;

            if (release == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }
            
            if (release.Exercise == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                        "exercise not found")));
                return;
            }
            
            if (release.Exercise.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "exercise is permanently locked")));
                return;
            }

            if (!await HasPermission(release.Id)) return;

            var userReleaseParticipation = await _context.ExerciseReleaseWithUserAsParticipations.Where(p =>
                        p.ExerciseReleaseId == assessmentFullBase.ExerciseReleaseId
                        && p.UserId == assessmentFullBase.UserId
                    )
                    .Include(p => p.Solutions)
                    .ThenInclude(p => p.Assessment)
                    .Include(p => p.User)
                    .FirstOrDefaultAsync()
                ;

            if (userReleaseParticipation == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no participation found")));
                return;
            }
           

            var solution =
                userReleaseParticipation.Solutions.FirstOrDefault(p => p.PLangId == assessmentFullBase.PLangId);

            if (solution == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                            "no solution for plang " + assessmentFullBase.PLangId + " found")));
                return;
            }

            if (solution.Assessment == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                            "automatic assessment has not finished yet and would overwrite values")));
                return;
            }

            //don't allow all props to be set... e.g. automatic points should not be altered manually
            solution.Assessment.NoteForOtherTutors = assessmentFullBase.NoteForOtherTutors;
            solution.Assessment.ManualPoints = assessmentFullBase.ManualPoints;
            solution.Assessment.FeedbackForStudent = assessmentFullBase.FeedbackForStudent;

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }
    }
}
