﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Threading.Tasks;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models;
using ClientServer.Models.Users;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;

namespace ClientServer.Controllers.Core.Users
{
    [Route(Constants.ApiPrefix + "users")]
    public class UsersController : ControllerWithDb
    {
        //frontend also hashes...
        public static string EmptyStringPw = PasswordHelper.HashLikeFrontend("");
            
        public UsersController(YapexDbContext context) : base(context)
        {
        }

        #region user data

        /// <summary>
        /// returns the user data for an user (from id)
        /// </summary>
        /// <param name="userId">the user id</param>
        /// <param name="context">the db context</param>
        /// <returns></returns>
        internal static async Task<UserDataFullBase> _GetUserData(int userId, YapexDbContext context)
        {
            User matchedUser = await context.Users
                .Include(p => p.SystemRole)
                .Include(p => p.SystemRole.SystemRolePermission)
                .Include(p => p.UserSettings)
                .Include(p => p.UserSettings.Lang)
                .Include(p => p.UserSettings.CodeEditorSetting)
                .FirstOrDefaultAsync(p => p.Id == userId);


            //user needs at least one group with create exercise permission...
            UserWithUserGroup groupConnection = await context.UserWithUserGroups
                .FirstOrDefaultAsync(p => p.UserId == userId
                                     && p.GroupRole.GroupRolePermission.CanCreateExercises);

            if (matchedUser == null)
            {
                return null;
            }


            var userData = new UserDataFullBase()
            {
                Id = matchedUser.Id,
                Email = matchedUser.Email,
                LastName = matchedUser.LastName,
                FirstName = matchedUser.FirstName,
                SystemRole = null,
                Settings = null,
                TempCanCreateExercise = (groupConnection != null)
            };

            if (matchedUser.SystemRole != null)
            {
                var role = matchedUser.SystemRole;

                userData.SystemRole = new SystemRoleFullBase()
                {
                    Id = role.Id,
                    DisplayName = role.DisplayName,
                    Email = role.Email,
                    CanChangeRoles = role.SystemRolePermission.CanChangeRoles,
                    CanCreateGroups = role.SystemRolePermission.CanCreateGroups,
                    CanDeleteGroups = role.SystemRolePermission.CanDeleteGroups,
                    CanCreateRoles = role.SystemRolePermission.CanCreateRoles,
                    CanDeleteRoles = role.SystemRolePermission.CanDeleteRoles,
                    CanDeleteActivatedUsers = role.SystemRolePermission.CanDeleteActivatedUsers,
                    CanManageNewUsers = role.SystemRolePermission.CanManageNewUsers,
                    CanChangeOtherUsersSystemRole = role.SystemRolePermission.CanChangeOtherUsersSystemRole,
                    CanChangeUserData = role.SystemRolePermission.CanChangeUserData,
                    CanManageTags = role.SystemRolePermission.CanManageTags,
                    CanChangeSystemSettings = role.SystemRolePermission.CanChangeSystemSettings,
                    CanViewDashboard = role.SystemRolePermission.CanViewDashboard
                };
            }

            if (matchedUser.UserSettings != null)
            {

                Lang lang = await context.Langs.FirstOrDefaultAsync();
                
                userData.Settings = new UserSettingsFullBase()
                {
                    Id = matchedUser.UserSettings.Id,
                    LangId = matchedUser.UserSettings.LangId ?? lang.Id,
                    LangShortcut = matchedUser.UserSettings.Lang.LangShortcut,
                    CodeEditorSettings = null,
                    Theme = matchedUser.UserSettings.Theme
                };

                if (matchedUser.UserSettings.CodeEditorSetting != null)
                {
                    userData.Settings.CodeEditorSettings = new CodeEditorSettingsFullBase()
                    {
                        Id = matchedUser.UserSettings.CodeEditorSetting.Id,
                        ShowLineIndentions = matchedUser.UserSettings.CodeEditorSetting.ShowLineIndentions,
                        ShowLineNumbers = matchedUser.UserSettings.CodeEditorSetting.ShowLineNumbers,
                        TabSize = matchedUser.UserSettings.CodeEditorSetting.TabSize,
                        ShowInvisibles = matchedUser.UserSettings.CodeEditorSetting.ShowInvisibles,
                        FontSize = matchedUser.UserSettings.CodeEditorSetting.FontSize,
                        Theme = matchedUser.UserSettings.CodeEditorSetting.Theme,
                        HighlightCurrentLine = matchedUser.UserSettings.CodeEditorSetting.HighlightCurrentLine,
                        UseWrapping = matchedUser.UserSettings.CodeEditorSetting.UseWrapping,
                    };
                }
            }


            return userData;
        }


        /// <summary>
        /// returns the user data if the user is logged in else not
        /// </summary>
        [HttpGet("data")]
        public async Task GetUserData()
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            User user = await _context.Users
                .FirstOrDefaultAsync(p => p.Id == userId);

            if (user == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user not found")));
                return;
            }

            UserDataFullBase userData = await _GetUserData(user.Id, this._context);

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<UserDataFullBase>(ResponseCode.Ok, "user data queried",
                        userData)));
        }

        /// <summary>
        /// updates the basic user data for the given user (by id)
        /// </summary>
        /// <param name="data">hte user data with the user id</param>
        [HttpPost("data/force")]
        public async Task AdminUpdateUserData([FromBody] BaseUserData data)
        {
            //only admins are allowed to change the user data or the user itself...

            if (!await base.IsLoggedIn()) return;


            if (
                !await HasSystemPermission(
                    (permission) => permission != null && permission.SystemRolePermission.CanChangeUserData))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            var user = await _context.Users
                .FirstOrDefaultAsync(p => p.Id == data.TargetUserId);

            if (user == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user not found")));
                return;
            }


            if (string.IsNullOrWhiteSpace(data.FirstName) == false)
            {
                user.FirstName = data.FirstName.Trim();
            }

            if (string.IsNullOrWhiteSpace(data.LastName) == false)
            {
                user.LastName = data.LastName.Trim();
            }

            if (string.IsNullOrWhiteSpace(data.Email) == false)
            {
                user.Email = data.Email.Trim();
            }

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
               await base.HandleDbError(ex);
                return;
            }


            BaseUserData responseObj = new BaseUserData()
            {
                Email = user.Email,
                LastName = user.LastName,
                FirstName = user.FirstName
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<BaseUserData>(ResponseCode.Ok, "ok", responseObj)));
        }

        /// <summary>
        /// resets the user password (if a given password is provided (not null or whitespaces) then this password is used instead of a random one)
        /// hte password is then returned
        /// </summary>
        /// <param name="tuple">the tuple with the target user id (and the new password)</param>
        [HttpPost("password/reset")]
        public async Task ResetPassword([FromBody] IdAndPasswordTuple tuple)
        {
            //only admins are allowed to change the user data or the user itself...

            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            if (
                !await HasSystemPermission(
                    (permission) => permission != null && permission.SystemRolePermission.CanChangeUserData))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            if (userId == tuple.TargetUserId)
            {
                //user cannot reset own password through this method...
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no allowed")));
                return;
            }

            var user = await _context.Users
                .FirstOrDefaultAsync(p => p.Id == tuple.TargetUserId);

            if (user == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user not found")));
                return;
            }


            string randomPassword;

            //not used yet because we always want the server to create a random password (in this version)
            tuple.Password = "";

            if (string.IsNullOrWhiteSpace(tuple.Password))
            {
                //empty --> generate new one
                randomPassword = PasswordHelper.GetRandomPassword();
            }
            else
            {
                //admin forces the given password
                randomPassword = tuple.Password;
            }

            //frontend users sha pre encryption... we need to use this too! because there is no good sha3 algorithm (sha512) for c# we use sha512...
            user.Password = PasswordHelper.GetHash(PasswordHelper.HashLikeFrontend(randomPassword));


            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
               await base.HandleDbError(ex);
                return;
            }


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<string>(ResponseCode.Ok, "ok", randomPassword)));
        }

        /// <summary>
        /// updates the basic user data for the current user
        /// </summary>
        /// <param name="data"></param>
        [HttpPost("data")]
        public async Task UpdateUserData([FromBody] EmailAndPasswordTuple data)
        {
            //user can only change email and password... other basic information can only be changed by admin...

            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            var user = await _context.Users
                .FirstOrDefaultAsync(p => p.Id == userId);

            if (user == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user not found")));
                return;
            }


            //ony set email if needed
            if (string.IsNullOrWhiteSpace(data.Email) == false)
            {
                user.Email = data.Email.Trim();
            }

            
            if (data.NewPassword != EmptyStringPw &&
                data.OldPassword != EmptyStringPw)
            {
                //check if user knowns his/her own password

                if (PasswordHelper.Verify(data.OldPassword, user.Password))
                {
                    var hashedPw = PasswordHelper.GetHash(data.NewPassword);
                    user.Password = hashedPw;
                }
                else
                {
                    //password was invalid...

                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "wrong old password")));
                    return;
                }
            }


            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
               await base.HandleDbError(ex);
                return;
            }


            BaseUserData responseObj = new BaseUserData()
            {
                Email = user.Email,
                LastName = user.LastName,
                FirstName = user.FirstName
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<BaseUserData>(ResponseCode.Ok, "ok", responseObj)));
        }

        #endregion

        #region new (not activated) users

        /// <summary>
        /// returns users that are not active yet
        /// </summary>
        [HttpPost("notactivated")]
        public async Task GetNotActivatedUsers([FromBody] PaginationPostData pagination)
        {
            if (!await base.IsLoggedIn()) return;


            if (
                !await HasSystemPermission(
                    (permission) => permission != null && permission.SystemRolePermission.CanManageNewUsers))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            int userId = GetUserId();

            var notActivatedUsersList = _context.Users
                    .Where((p) => p.IsActivated == false)
                ;
            
            if (String.IsNullOrWhiteSpace(pagination.SearchText) == false)
            {
                var text = pagination.SearchText.ToLower();
                notActivatedUsersList = notActivatedUsersList
                    .Where(p => (p.Email != null && p.Email.ToLower().Contains(text))
                                || (p.Token != null && p.Token.ToLower().Contains(text))
                                || (p.FirstName != null && p.FirstName.ToLower().Contains(text))
                                || (p.LastName != null && p.LastName.ToLower().Contains(text))
                                || (p.SystemRole != null && p.SystemRole.DisplayName != null &&
                                    p.SystemRole.DisplayName.ToLower().Contains(text))
                    );
            }

            string orderByProp = "";
            Expression<Func<User, object>> orderByFunc = null;

            //not all props are sortable via the frontend
            if (String.IsNullOrWhiteSpace(pagination.OrderByProp) == false)
            {
                if (pagination.OrderByProp.ToLower() == "id".ToLower())
                {
                    orderByFunc = p => p.Id;
                    orderByProp = "id".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "lastName".ToLower())
                {
                    orderByFunc = p => p.LastName;
                    orderByProp = "lastname";
                }
                else if (pagination.OrderByProp.ToLower() == "firstName".ToLower())
                {
                    orderByFunc = p => p.FirstName;
                    orderByProp = "firstname";
                }
                else if (pagination.OrderByProp.ToLower() == "token".ToLower())
                {
                    orderByFunc = p => p.Token;
                    orderByProp = "token";
                }
                else if (pagination.OrderByProp.ToLower() == "email".ToLower())
                {
                    orderByFunc = p => p.Email;
                    orderByProp = "email";
                }
//                else if (paginationPostData.OrderByProp.ToLower() == "systemRole".ToLower())
//                {
//                    orderByFunc = p => p.SystemRole == null ? "" : p.SystemRole.DisplayName;
//                    orderByProp = "systemrole";
//                }
                else if (pagination.OrderByProp.ToLower() == "createdAt".ToLower())
                {
                    orderByFunc = p => p.CreatedAt;
                    orderByProp = "createdat";
                }
//                else if (paginationPostData.OrderByProp.ToLower() == "LastLoginAt".ToLower())
//                {
//                    orderByFunc = p => p.LastLoginAt;
//                    orderByProp = "LastLoginAt".ToLower();
//                }
                else if (pagination.OrderByProp.ToLower() == "IsActivated".ToLower())
                {
                    orderByFunc = p => p.IsActivated;
                    orderByProp = "IsActivated".ToLower();
                }
            }

            var paginatedData = await PaginationHelper.GetAsPaginationAsync(notActivatedUsersList,
                pagination.Page, pagination.PageSize, pagination.IsOrderByDesc,
                orderByFunc, orderByProp, pagination.SearchText
            );

            List<UserPreviewFromBackend> frontendUsers = TransformUsersForFrontend(paginatedData.Items, true, userId);

            var frontendData = PaginationHelper.TransformForFrontend(paginatedData, frontendUsers);

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<PaginatedData<UserPreviewFromBackend>>(ResponseCode.Ok,
                        "not activated user queried", frontendData)));
        }

        /// <summary>
        /// deletes a not actiavted user
        /// </summary>
        /// <param name="userId">the user id</param>
        [HttpDelete("notactivated/{userId}")]
        public async Task DeleteNewUser(int userId)
        {
            if (!await base.IsLoggedIn()) return;

            if (
                    !await HasSystemPermission(
                        permission => permission != null && permission.SystemRolePermission.CanManageNewUsers))
                //TODo or use permission can delete users??
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            User user = await _context.Users.FirstOrDefaultAsync(p => p.Id == userId && p.IsActivated == false);

            if (user == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user not found")));
                return;
            }

            try
            {
                _context.Users.Remove(user);
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
               await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "not activated user deleted")));
        }


        /// <summary>
        /// activates the given user and assign the given system role
        /// </summary>
        /// <param name="tuple">the user and system role</param>
        [HttpPost("activate")]
        public async Task SetUserActive([FromBody] ActivateUserTuple tuple)
        {
            if (!await base.IsLoggedIn()) return;

            if (
                !await HasSystemPermission(
                    permission => permission != null && permission.SystemRolePermission.CanManageNewUsers))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            User user =
                await _context.Users.FirstOrDefaultAsync(p => p.Id == tuple.UserId && p.IsActivated == false);

            if (user == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user not found")));
                return;
            }

            //set user as active
            user.IsActivated = true;

            if (tuple.SystemRoleId != null)
            {
                //check if the creating user is allowed to set the role...
                if (
                    !await HasSystemPermission(
                        permission =>
                            permission != null && permission.SystemRolePermission.CanChangeOtherUsersSystemRole))
                {
                    //TODO MAYBE not fail here jsut drop the system role?
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NoPermission,
                                "no permission to set system role")));
                    return;
                }

                if (tuple.SystemRoleId > 0)
                {
                    //check if system role exists
                    SystemRole systemRole = await _context.SystemRoles.FirstOrDefaultAsync(p => p.Id == tuple.SystemRoleId);

                    if (systemRole == null)
                    {
                        await
                            Response.WriteAsync(
                                Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system role not found")));
                        return;
                    }
                }

                //set the system role
                user.SystemRoleId = tuple.SystemRoleId;
            }


            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
               await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "user activated")));
        }

        #endregion

        #region activated users

        /// <summary>
        /// returns all activated users
        /// </summary>
        [HttpPost("activated")]
        public async Task GetActivatedUsers([FromBody] PaginationPostData pagination)
        {
            if (!await base.IsLoggedIn()) return;

            if (
                !await HasSystemPermission(
                    (permission) => permission != null
                                    && (permission.SystemRolePermission.CanChangeOtherUsersSystemRole
                                        || permission.SystemRolePermission.CanDeleteActivatedUsers
                                    )))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            int userId = GetUserId();
            
            var activatedUsersList = _context.Users
                    .Where((p) => p.IsActivated == true)
                ;

            if (String.IsNullOrWhiteSpace(pagination.SearchText) == false)
            {
                var text = pagination.SearchText.ToLower();
                activatedUsersList = activatedUsersList
                    .Where(p => (p.Email != null && p.Email.ToLower().Contains(text))
                                || (p.Token != null && p.Token.ToLower().Contains(text))
                                || (p.FirstName != null && p.FirstName.ToLower().Contains(text))
                                || (p.LastName != null && p.LastName.ToLower().Contains(text))
                                || (p.SystemRole != null && p.SystemRole.DisplayName != null &&
                                    p.SystemRole.DisplayName.ToLower().Contains(text))
                    );
            }

            string orderByProp = "";
            Expression<Func<User, object>> orderByFunc = null;

            if (String.IsNullOrWhiteSpace(pagination.OrderByProp) == false)
            {
                if (pagination.OrderByProp.ToLower() == "id".ToLower())
                {
                    orderByFunc = p => p.Id;
                    orderByProp = "id".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "lastName".ToLower())
                {
                    orderByFunc = p => p.LastName;
                    orderByProp = "lastname";
                }
                else if (pagination.OrderByProp.ToLower() == "firstName".ToLower())
                {
                    orderByFunc = p => p.FirstName;
                    orderByProp = "firstname";
                }
                else if (pagination.OrderByProp.ToLower() == "token".ToLower())
                {
                    orderByFunc = p => p.Token;
                    orderByProp = "token";
                }
                else if (pagination.OrderByProp.ToLower() == "email".ToLower())
                {
                    orderByFunc = p => p.Email;
                    orderByProp = "email";
                }
                else if (pagination.OrderByProp.ToLower() == "systemRole".ToLower())
                {
                    orderByFunc = p => p.SystemRole == null ? "" : p.SystemRole.DisplayName;
                    orderByProp = "systemrole";
                }
                else if (pagination.OrderByProp.ToLower() == "createdAt".ToLower())
                {
                    orderByFunc = p => p.CreatedAt;
                    orderByProp = "createdat";
                }
                else if (pagination.OrderByProp.ToLower() == "LastLoginAt".ToLower())
                {
                    orderByFunc = p => p.LastLoginAt;
                    orderByProp = "LastLoginAt".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "IsActivated".ToLower())
                {
                    orderByFunc = p => p.IsActivated;
                    orderByProp = "IsActivated".ToLower();
                }
            }

            var paginatedActivatedUsersList = await PaginationHelper.GetAsPaginationAsync(activatedUsersList,
                pagination.Page, pagination.PageSize, pagination.IsOrderByDesc,
                orderByFunc, orderByProp, pagination.SearchText
            );


            List<UserPreviewFromBackend> activeUsersForFrontend = TransformUsersForFrontend(paginatedActivatedUsersList.Items, true, userId);

            var frontendData = PaginationHelper.TransformForFrontend(paginatedActivatedUsersList, activeUsersForFrontend);

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<PaginatedData<UserPreviewFromBackend>>(ResponseCode.Ok,
                        "active user queried", frontendData)));
        }


        /// <summary>
        /// deletes an activated user
        /// a user cannot delete him/herself!!
        /// </summary>
        /// <param name="userId"></param>
        [HttpDelete("activated/{userId}")]
        public async Task DeleteOldUser(int userId)
        {
            if (!await base.IsLoggedIn()) return;

            int ownId = GetUserId();

            if (
                !await HasSystemPermission(
                    permission => permission != null && permission.SystemRolePermission.CanDeleteActivatedUsers))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            if (userId == ownId)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "cannot delete user (yourself)")));
                return;
            }

            User user = await _context.Users.FirstOrDefaultAsync(p => p.Id == userId && p.IsActivated);

            if (user == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user not found")));
                return;
            }

            //TODO MUST check if all custom tests and exercise solutions and group memberships gets deleted!!

            var participationWhereExIsPermanentlyLockedCount = await _context.ExerciseReleaseWithUserAsParticipations
                .CountAsync(p => p.UserId == userId && p.ExerciseRelease.Exercise.IsPermanentlyLocked)
                ;

            if (participationWhereExIsPermanentlyLockedCount > 0)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "cannot delete user because he has participation(s) (solutions) für locked exercise(s)")));
                return;
            }

            
            //delete all exercises where the user is thw owner but the exercise is hidden... (because no one has access to this exercises)

            var hiddenExercises = await _context.Exercises
                    .Where(p => p.UserId == userId && p.IsOnlyVisibleToOwner)
                    .ToListAsync()
                ;
            
            _context.Exercises.RemoveRange(hiddenExercises);
            

            try
            {
                _context.Users.Remove(user);
                await _context.SaveChangesAsync();

            }
            catch (Exception ex)
            {
               await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "new user deleted")));
        }


        //this can also be used to set the role to null (take system role id 0 or -1?)
        /// <summary>
        /// sets the system role for an user
        /// </summary>
        /// <param name="tuple">the system role and the user</param>
        [HttpPost("activated/systemrole")]
        public async Task SetUserSystemRole([FromBody] ActivateUserTuple tuple)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            if (
                !await HasSystemPermission(
                    permission => permission != null && permission.SystemRolePermission.CanChangeOtherUsersSystemRole))
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            //TODO MAYBE user cannot change own role (he/she could drop all privileges)
            if (userId == tuple.UserId)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "cannot set own role")));
                return;
            }


            if (userId == tuple.UserId && tuple.SystemRoleId == null)
            {
                //user cant remove his role...
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "cannot set role to null (yourself)")));
                return;
            }

            if (tuple.SystemRoleId != null)
            {
                //check if system role exists
                SystemRole systemRole = await _context.SystemRoles.FirstOrDefaultAsync(p => p.Id == tuple.SystemRoleId);

                if (systemRole == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system role not found")));
                    return;
                }
            }


            User user =
                await _context.Users.FirstOrDefaultAsync(p => p.Id == tuple.UserId && p.IsActivated);

            if (user == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user not found")));
                return;
            }


            user.SystemRoleId = tuple.SystemRoleId;

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
               await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "role changed")));
        }

        #endregion

        public List<UserPreviewFromBackend> TransformUsersForFrontend(List<User> users, bool areActivatedUsers, int currentUserId)
        {
            var tempUsers = new List<UserPreviewFromBackend>();

            foreach (var user in users)
            {
                var userPreviewFromBackend = new UserPreviewFromBackend()
                {
                    Id = user.Id,
                    Token = user.Token,
                    FirstName = user.FirstName,
                    LastName = user.LastName,
                    Email = user.Email,
                    IsActivated = user.IsActivated,
                    IsCurrentUser = (user.Id == currentUserId),
                    SystemRoleId = user.SystemRoleId,
                    CreatedAt = user.CreatedAt
                };

                if (areActivatedUsers)
                {
                    userPreviewFromBackend.LastLoginAt = user.LastLoginAt;
                }

                tempUsers.Add(userPreviewFromBackend);
            }

            return tempUsers;
        }

        /// <summary>
        /// a user for the frontend (to display in a list)
        /// </summary>
        public class UserPreviewFromBackend
        {
            public int Id { get; set; }

            /// <summary>
            /// the token (id)
            /// </summary>
            public string Token { get; set; }

            /// <summary>
            /// the first name
            /// </summary>
            public string FirstName { get; set; }

            /// <summary>
            /// the last name
            /// </summary>
            public string LastName { get; set; }

            /// <summary>
            /// the email
            /// </summary>
            public string Email { get; set; }

            /// <summary>
            /// true: user is activated and can login, fale: not
            /// </summary>
            public bool IsActivated { get; set; }

            /// <summary>
            /// the system role id or null
            /// </summary>
            public int? SystemRoleId { get; set; }

            /// <summary>
            /// the date time when the user did a login (frontend this is string ... why?)
            /// </summary>
            public DateTime? LastLoginAt { get; set; }

            /// <summary>
            /// the date time when the user was created (might be helpful for the new users list)
            /// </summary>
            public DateTime? CreatedAt { get; set; }

            /// <summary>
            /// this is calculated, not stored
            /// </summary>
            public bool IsCurrentUser { get; set; }
        }

        /// <summary>
        /// a tuple to activate a user / or change a users system role
        /// </summary>
        public class ActivateUserTuple
        {
            /// <summary>
            /// the user id
            /// </summary>
            public int UserId { get; set; }

            /// <summary>
            /// the system role id or null
            /// </summary>
            public int? SystemRoleId { get; set; }
        }
    }


    /// <summary>
    /// a tuple to reset a users password
    /// </summary>
    public class IdAndPasswordTuple
    {
        public int TargetUserId { get; set; }

        /// <summary>
        /// the password, CAN BE NULL to generate a random password
        /// </summary>
        public string Password { get; set; }
    }

    /// <summary>
    /// tuple to change the user data (changed by user)
    /// </summary>
    public class EmailAndPasswordTuple
    {
        /// <summary>
        /// the email
        /// </summary>
        public string Email { get; set; }

        /// <summary>
        /// the old password
        /// </summary>
        public string OldPassword { get; set; }

        /// <summary>
        /// the new user password (needs to be SHA 512 encrypted from the frontend)
        /// </summary>
        public string NewPassword { get; set; }
    }
}
